2#include "config.h"

__kernel void matmul(__global T* A, __global T* B, __global T* C, int n) {

    if (get_global_id(0) >= n || get_global_id(1) >= n) return;

    T sum = 0;

    // get number of groups (same on both dimensions)
    int num_groups = get_num_groups(0);

    // declaration of the local memory array As 
    // used to store the sub-matrix of A
    __local T As[LOCAL_SIZE][LOCAL_SIZE];
    
    // declaration of the local memory array Bs 
    // used to store the sub-matrix of B
    __local T Bs[LOCAL_SIZE][LOCAL_SIZE];

    // group indices in MDRange
    int ig = get_group_id(1);
    int jg = get_group_id(0);

    // local indices in work group
    int i = get_local_id(1);
    int j = get_local_id(0);

    // loop over all the sub-matrices of A and B
    // required to compute the block sub-matrix
    for (int kg = 0; kg < num_groups; kg++) {

        // load the matrices from global memory
        // to local memory; each thread loads
        // one element of each matrix
        As[i][j] = A[(ig*LOCAL_SIZE+i)*n + kg*LOCAL_SIZE+j];
        Bs[i][j] = B[(kg*LOCAL_SIZE+i)*n + jg*LOCAL_SIZE+j];

        // synchronize to make sure matrices are loaded
        barrier(CLK_LOCAL_MEM_FENCE);
        
        // multiply the two matrices together;
        // each thread computes one element
        // of the block sub-matrix
        for (int k = 0; k < LOCAL_SIZE; k++)
            sum += As[i][k] * Bs[k][j];

        // synchronize to make sure that the preceding
        // computation is done before loading two new
        // sub-matrices of A and B in the next iteration
        barrier(CLK_LOCAL_MEM_FENCE);
    }

    // write the block sub-matrix to device global memory;
    // each thread writes one element
    C[(ig*LOCAL_SIZE+i)*n + jg*LOCAL_SIZE+j] = sum;
}
