#include "config.h"

__kernel void matmul(__global T* A, __global T* B, __global T* C, int n) {

    if (get_global_id(0) >= n || get_global_id(1) >= n) return;

    T sum = 0;

    // start indices of work group
    int y = get_group_id(1)*LOCAL_SIZE;
    int x = get_group_id(0)*LOCAL_SIZE;

    // local indices in work group
    int u = get_local_id(1);
    int v = get_local_id(0);

    // loop over all the sub-matrices of A and B
    // required to compute the block sub-matrix
    for (int z = 0; z < n; z += LOCAL_SIZE) {

        // declaration of the local memory array As 
        // used to store the sub-matrix of A
        __local T As[LOCAL_SIZE][LOCAL_SIZE];
        
        // declaration of the local memory array Bs 
        // used to store the sub-matrix of B
        __local T Bs[LOCAL_SIZE][LOCAL_SIZE];

        // load the matrices from global memory
        // to local memory; each thread loads
        // one element of each matrix
        As[u][v] = A[(y+u)*n + z+v];
        Bs[u][v] = B[(z+v)*n + x+v];

        // synchronize to make sure matrices are loaded
        barrier(CLK_LOCAL_MEM_FENCE);

        // multiply the two matrices together;
        // each thread computes one element
        // of the block sub-matrix
        for (int w = 0; w < LOCAL_SIZE; w++)
            sum += As[u][w] * Bs[w][v];

        // synchronize to make sure that the preceding
        // computation is done before loading two new
        // sub-matrices of A and B in the next iteration
        barrier(CLK_LOCAL_MEM_FENCE);
    }

    // write the block sub-matrix to device global memory;
    // each thread writes one element
    C[i*n+j] = sum;
}
